﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  /// <summary>
  /// An operator for analyzing the best solution of Generalized Quadratic Assignment Problems.
  /// </summary>
  [Item("Best GeneralizedQAP Solution Analyzer", "An operator for analyzing the best solution of Generalized Quadratic Assignment Problems.")]
  [StorableClass]
  public sealed class BestGQAPSolutionAnalyzer : SingleSuccessorOperator, IAssignmentsAwareGQAPOperator,
    IQualitiesAwareGQAPOperator, IDistancesAwareGQAPOperator, IWeightsAwareGQAPOperator, IInstallationCostsAwareGQAPOperator,
    IDemandsAwareGQAPOperator, ICapacitiesAwareGQAPOperator, ITransportationCostsAwareGQAPOperator,
    IExpectedRandomQualityAwareGQAPOperator, IEquipmentNamesAwareGQAPOperator, ILocationNamesAwareGQAPOperator,
    IBestKnownQualityAwareGQAPOperator, IBestKnownSolutionAwareGQAPOperator,
    IEvaluatorAwareGQAPOperator, IAnalyzer {

    public bool EnabledByDefault {
      get { return true; }
    }

    #region Parameter Properties
    public IScopeTreeLookupParameter<IntegerVector> AssignmentParameter {
      get { return (IScopeTreeLookupParameter<IntegerVector>)Parameters["Assignment"]; }
    }
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> FlowDistanceQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["FlowDistanceQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> InstallationQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["InstallationQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> OverbookedCapacityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["OverbookedCapacity"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public IValueLookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public IValueLookupParameter<DoubleValue> ExpectedRandomQualityParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["ExpectedRandomQuality"]; }
    }
    public ILookupParameter<StringArray> EquipmentNamesParameter {
      get { return (ILookupParameter<StringArray>)Parameters["EquipmentNames"]; }
    }
    public ILookupParameter<StringArray> LocationNamesParameter {
      get { return (ILookupParameter<StringArray>)Parameters["LocationNames"]; }
    }
    public ILookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public ILookupParameter<GQAPAssignment> BestKnownSolutionParameter {
      get { return (ILookupParameter<GQAPAssignment>)Parameters["BestKnownSolution"]; }
    }
    public ILookupParameter<GQAPAssignment> BestSolutionParameter {
      get { return (ILookupParameter<GQAPAssignment>)Parameters["BestSolution"]; }
    }
    public IValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (IValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public IValueLookupParameter<IGQAPEvaluator> EvaluatorParameter {
      get { return (IValueLookupParameter<IGQAPEvaluator>)Parameters["Evaluator"]; }
    }
    #endregion

    [StorableConstructor]
    private BestGQAPSolutionAnalyzer(bool deserializing) : base(deserializing) { }
    private BestGQAPSolutionAnalyzer(BestGQAPSolutionAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestGQAPSolutionAnalyzer(this, cloner);
    }
    public BestGQAPSolutionAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", GeneralizedQuadraticAssignmentProblem.MaximizationDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", GeneralizedQuadraticAssignmentProblem.DistancesDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", GeneralizedQuadraticAssignmentProblem.WeightsDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", GeneralizedQuadraticAssignmentProblem.InstallationCostsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", GeneralizedQuadraticAssignmentProblem.DemandsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", GeneralizedQuadraticAssignmentProblem.CapacitiesDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("TransportationCosts", GeneralizedQuadraticAssignmentProblem.TransportationCostsDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("ExpectedRandomQuality", GeneralizedQuadraticAssignmentProblem.ExpectedRandomQualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<IntegerVector>("Assignment", GQAPSolutionCreator.AssignmentDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", GQAPEvaluator.QualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("FlowDistanceQuality", GQAPEvaluator.FlowDistanceQualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("InstallationQuality", GQAPEvaluator.InstallationQualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("OverbookedCapacity", GQAPEvaluator.OverbookedCapacityDescription));
      Parameters.Add(new LookupParameter<StringArray>("EquipmentNames", GeneralizedQuadraticAssignmentProblem.EquipmentNamesDescription));
      Parameters.Add(new LookupParameter<StringArray>("LocationNames", GeneralizedQuadraticAssignmentProblem.LocationNamesDescription));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", GeneralizedQuadraticAssignmentProblem.BestKnownQualityDescription));
      Parameters.Add(new LookupParameter<GQAPAssignment>("BestKnownSolution", GeneralizedQuadraticAssignmentProblem.BestKnownSolutionDescription));
      Parameters.Add(new LookupParameter<GQAPAssignment>("BestSolution", "The best GQAP solution found so far."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "The result collection where the best GQAP solution should be stored."));
      Parameters.Add(new ValueLookupParameter<IGQAPEvaluator>("Evaluator", "The evaluator that is used to evaluate GQAP solutions."));
    }

    public override IOperation Apply() {
      var evaluator = EvaluatorParameter.ActualValue;
      var assignments = AssignmentParameter.ActualValue;
      var qualities = QualityParameter.ActualValue;
      var equipmentNames = EquipmentNamesParameter.ActualValue;
      var locationNames = LocationNamesParameter.ActualValue;
      var flowDistanceQualities = FlowDistanceQualityParameter.ActualValue;
      var installationQualities = InstallationQualityParameter.ActualValue;
      var overbookedCapacities = OverbookedCapacityParameter.ActualValue;
      var distances = DistancesParameter.ActualValue;
      var weights = WeightsParameter.ActualValue;
      var installationCosts = InstallationCostsParameter.ActualValue;
      var demands = DemandsParameter.ActualValue;
      var capacities = CapacitiesParameter.ActualValue;
      var transportationCosts = TransportationCostsParameter.ActualValue;
      var expectedRandomQuality = ExpectedRandomQualityParameter.ActualValue;
      var results = ResultsParameter.ActualValue;
      var maximization = MaximizationParameter.ActualValue.Value;
      var bestKnownQuality = BestKnownQualityParameter.ActualValue;

      int bestIndex;
      var tmp = qualities.Select((x, index) => new { Index = index, Value = x.Value });
      if (maximization) bestIndex = tmp.MaxItems(x => x.Value).First().Index;
      else bestIndex = tmp.MinItems(x => x.Value).First().Index;

      if (bestKnownQuality == null || HasSolutionImproved(bestKnownQuality.Value, qualities[bestIndex].Value, maximization)) {
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[bestIndex].Value);
        BestKnownSolutionParameter.ActualValue = new GQAPAssignment((IntegerVector)assignments[bestIndex].Clone(), (DoubleValue)qualities[bestIndex].Clone(), flowDistanceQualities[bestIndex], installationQualities[bestIndex], overbookedCapacities[bestIndex], equipmentNames, locationNames, distances, weights, installationCosts, demands, capacities, transportationCosts, expectedRandomQuality, evaluator);
      }

      var assignment = BestSolutionParameter.ActualValue;
      if (assignment == null) {
        assignment = new GQAPAssignment((IntegerVector)assignments[bestIndex].Clone(), (DoubleValue)qualities[bestIndex].Clone(), flowDistanceQualities[bestIndex], installationQualities[bestIndex], overbookedCapacities[bestIndex], equipmentNames, locationNames, distances, weights, installationCosts, demands, capacities, transportationCosts, expectedRandomQuality, evaluator);
        assignment.Distances = distances;
        BestSolutionParameter.ActualValue = assignment;
      } else {
        if (HasSolutionImproved(assignment.Quality.Value, qualities[bestIndex].Value, maximization)) {
          assignment.Assignment = (IntegerVector)assignments[bestIndex].Clone();
          assignment.Quality = (DoubleValue)qualities[bestIndex].Clone();
          assignment.FlowDistanceQuality = (DoubleValue)flowDistanceQualities[bestIndex].Clone();
          assignment.InstallationQuality = (DoubleValue)installationQualities[bestIndex].Clone();
          assignment.OverbookedCapacity = (DoubleValue)overbookedCapacities[bestIndex].Clone();
          assignment.EquipmentNames = equipmentNames;
          assignment.LocationNames = locationNames;
          assignment.Distances = distances;
          assignment.Weights = weights;
          assignment.InstallationCosts = installationCosts;
          assignment.Demands = demands;
          assignment.Capacities = capacities;
          assignment.TransportationCosts = transportationCosts;
          assignment.ExpectedRandomQuality = expectedRandomQuality;
          assignment.Evaluator = evaluator;
        }
      }
      if (!results.ContainsKey("Best GQAP Assignment"))
        results.Add(new Result("Best GQAP Assignment", assignment));

      return base.Apply();
    }

    private static bool HasSolutionImproved(double oldQuality, double quality, bool maximization) {
      return maximization && oldQuality < quality || !maximization && oldQuality > quality;
    }
  }
}
